import streamlit as st
from trulens.core import __package__ as core_package
from trulens.core import __version__ as core_version
from trulens.core.utils.imports import static_resource
from trulens.dashboard import __version__


def set_page_config(page_title="TruLens"):
    st.set_page_config(
        page_title=page_title,
        page_icon="https://www.trulens.org/img/favicon.ico",
        layout="wide",
    )

    logo = static_resource("dashboard", "ux/trulens_logo.svg")

    st.logo(str(logo), link="https://www.trulens.org/")

    st.markdown(
        """
        <style>
            [data-testid="stSidebarNav"] {
                background-repeat: no-repeat;
                background-size: 300px auto;
                background-position: 20px 20px;
                height: calc(100vh - 80px);
            }
            [data-testid="stSidebarNav"]::before {
                margin-left: 20px;
                margin-top: 20px;
                font-size: 30px;
                position: relative;
                top: 100px;
            }

            /* For user feedback button and version text */
            [data-testid="stSidebarUserContent"] {
                padding-bottom: 1rem;
            }

            [data-testid="stSidebarUserContent"] [data-testid="column"] {
                align-content: center;
            }

            [data-testid="stSidebarUserContent"] [data-testid="stText"] {
                color: #aaaaaa;
                font-size: 9pt;
            }

            /* For list items in st.dataframe */
            #portal .clip-region .boe-bubble {
                height: auto;
                border-radius: 4px;
                padding: 8px;
            }
        </style>
        """,
        unsafe_allow_html=True,
    )

    # Sidebar feedback button
    with st.sidebar:
        version_col, user_feedback_col = st.columns(2)
        with version_col:
            st.text(f"{core_package}\nv{core_version}")
            st.text(f"{__package__}\nv{__version__}")
        with user_feedback_col:
            st.link_button(
                "Share Feedback",
                "https://forms.gle/HAc4HBk5nZRpgw7C6",
                help="Help us improve TruLens!",
            )



from collections import defaultdict
from enum import Enum
import operator
from typing import Callable, List, NamedTuple, Optional

import numpy as np
from trulens.core.utils.serial import SerialModel


class ResultCategoryType(Enum):
    PASS = 0
    WARNING = 1
    FAIL = 2


class CATEGORY:
    """
    Feedback result categories for displaying purposes: pass, warning, fail, or
    unknown.
    """

    class Category(SerialModel):
        name: str
        adjective: str
        threshold: float
        color: str
        icon: str
        direction: Optional[str] = None
        compare: Optional[Callable[[float, float], bool]] = None

    class FeedbackDirection(NamedTuple):
        name: str
        ascending: bool
        thresholds: List[float]

    # support both directions by default
    # TODO: make this configurable (per feedback definition & per app?)
    directions = [
        FeedbackDirection("HIGHER_IS_BETTER", False, [0, 0.6, 0.8]),
        FeedbackDirection("LOWER_IS_BETTER", True, [0.2, 0.4, 1]),
    ]

    styling = {
        "PASS": dict(color="#aaffaa", icon="✅"),
        "WARNING": dict(color="#ffffaa", icon="⚠️"),
        "FAIL": dict(color="#ffaaaa", icon="🛑"),
    }

    PASS: defaultdict = defaultdict(dict)
    FAIL: defaultdict = defaultdict(dict)
    WARNING: defaultdict = defaultdict(dict)

    for direction in directions:
        a = sorted(
            zip(["low", "medium", "high"], sorted(direction.thresholds)),
            key=operator.itemgetter(1),
            reverse=not direction.ascending,
        )

        for enum, (adjective, threshold) in enumerate(a):
            category_name = ResultCategoryType(enum).name
            locals()[category_name][direction.name] = Category(
                name=category_name.lower(),
                adjective=adjective,
                threshold=threshold,
                direction=direction.name,
                compare=operator.ge
                if direction.name == "HIGHER_IS_BETTER"
                else operator.le,
                **styling[category_name],
            )

    UNKNOWN = Category(
        name="unknown",
        adjective="unknown",
        threshold=np.nan,
        color="#aaaaaa",
        icon="?",
    )

    # order matters here because `of_score` returns the first best category
    ALL = [PASS, WARNING, FAIL]  # not including UNKNOWN intentionally

    @staticmethod
    def of_score(score: float, higher_is_better: bool = True) -> Category:
        direction_key = (
            "HIGHER_IS_BETTER" if higher_is_better else "LOWER_IS_BETTER"
        )

        for cat in map(operator.itemgetter(direction_key), CATEGORY.ALL):
            if cat.compare(score, cat.threshold):
                return cat

        return CATEGORY.UNKNOWN


default_direction = "HIGHER_IS_BETTER"

# These would be useful to include in our pages but don't yet see a way to do
# this in streamlit.
root_js = f"""
    var default_pass_threshold = {CATEGORY.PASS[default_direction].threshold};
    var default_warning_threshold = {CATEGORY.WARNING[default_direction].threshold};
    var default_fail_threshold = {CATEGORY.FAIL[default_direction].threshold};
"""

# Not presently used. Need to figure out how to include this in streamlit pages.
root_html = f"""
<script>
    {root_js}
</script>
"""

stmetricdelta_hidearrow = """
    <style> [data-testid="stMetricDelta"] svg { display: none; } </style>
    """

valid_directions = ["HIGHER_IS_BETTER", "LOWER_IS_BETTER"]

cellstyle_jscode = {
    k: """function(params) {
        let v = parseFloat(params.value);
        """
    + "\n".join(
        f"""
        if (v {">=" if k == "HIGHER_IS_BETTER" else "<="} {cat.threshold}) {{
            return {{
                'color': 'black',
                'backgroundColor': '{cat.color}'
            }};
        }}
    """
        for cat in map(operator.itemgetter(k), CATEGORY.ALL)
    )
    + f"""
        // i.e. not a number
        return {{
            'color': 'black',
            'backgroundColor': '{CATEGORY.UNKNOWN.color}'
        }};
    }}"""
    for k in valid_directions
}

hide_table_row_index = """
    <style>
        thead tr th:first-child {display:none}
        tbody th {display:none}
    </style>
    """



import json
import random
from typing import Dict, List

import pandas as pd
import streamlit as st
from trulens.core.app import ComponentView
from trulens.core.schema.record import Record
from trulens.core.schema.record import RecordAppCall
from trulens.core.schema.select import Select
from trulens.core.schema.types import Metadata
from trulens.core.schema.types import Tags
from trulens.core.utils.constants import CLASS_INFO
from trulens.core.utils.containers import is_empty
from trulens.core.utils.json import jsonify
from trulens.core.utils.keys import REDACTED_VALUE
from trulens.core.utils.keys import should_redact_key
from trulens.core.utils.pyschema import is_noserio
from trulens.core.utils.serial import GetItemOrAttribute
from trulens.core.utils.serial import Lens


def write_or_json(st, obj):
    """
    Dispatch either st.json or st.write depending on content of `obj`. If it is
    a string that can parses into strictly json (dict), use st.json, otherwise
    use st.write.
    """

    if isinstance(obj, str):
        try:
            content = json.loads(obj)
            if not isinstance(content, str):
                st.json(content)
            else:
                st.write(content)

        except BaseException:
            st.write(obj)


def copy_to_clipboard(path, *args, **kwargs):
    st.session_state.clipboard = str(path)


def draw_selector_button(path) -> None:
    st.button(
        key=str(random.random()),
        label=f"{Select.render_for_dashboard(path)}",
        on_click=copy_to_clipboard,
        args=(path,),
    )


def render_selector_markdown(path) -> str:
    return f"[`{Select.render_for_dashboard(path)}`]"


def render_call_frame(frame: RecordAppCall, path=None) -> str:  # markdown
    path = path or frame.path

    return f"__{frame.method.name}__ (__{frame.method.obj.cls.module.module_name}.{frame.method.obj.cls.name}__)"


def dict_to_md(dictionary: dict) -> str:
    if len(dictionary) == 0:
        return "No metadata."
    mdheader = "|"
    mdseparator = "|"
    mdbody = "|"
    for key, value in dictionary.items():
        mdheader = mdheader + str(key) + "|"
        mdseparator = mdseparator + "-------|"
        mdbody = mdbody + str(value) + "|"
    mdtext = mdheader + "\n" + mdseparator + "\n" + mdbody
    return mdtext


def draw_metadata_and_tags(metadata: Metadata, tags: Tags) -> str:
    if isinstance(metadata, Dict):
        metadata["tags"] = tags
        return dict_to_md(metadata)
    else:
        return str(metadata) + "\n" + str(tags)


def draw_call(call: RecordAppCall) -> None:
    top = call.stack[-1]

    path = Select.for_record(
        top.path._append(
            step=GetItemOrAttribute(item_or_attribute=top.method.name)
        )
    )

    with st.expander(
        label="Call "
        + render_call_frame(top, path=path)
        + " "
        + render_selector_markdown(path)
    ):
        args = call.args
        rets = call.rets

        for frame in call.stack[::-1][1:]:
            st.write("Via " + render_call_frame(frame, path=path))

        st.subheader(f"Inputs {render_selector_markdown(path.args)}")
        if isinstance(args, Dict):
            st.json(args)
        else:
            st.write(args)

        st.subheader(f"Outputs {render_selector_markdown(path.rets)}")
        if isinstance(rets, Dict):
            st.json(rets)
        else:
            st.write(rets)


def draw_calls(record: Record, index: int) -> None:
    """
    Draw the calls recorded in a `record`.
    """

    calls = record.calls

    app_step = 0

    for call in calls:
        app_step += 1

        if app_step != index:
            continue

        draw_call(call)


def draw_prompt_info(query: Lens, component: ComponentView) -> None:
    prompt_details_json = jsonify(component.json, skip_specials=True)

    st.caption("Prompt details")

    path = Select.for_app(query)

    prompt_types = {
        k: v
        for k, v in prompt_details_json.items()
        if (v is not None)
        and not is_empty(v)
        and not is_noserio(v)
        and k != CLASS_INFO
    }

    for key, value in prompt_types.items():
        with st.expander(
            key.capitalize()
            + " "
            + render_selector_markdown(getattr(path, key)),
            expanded=True,
        ):
            if isinstance(value, (Dict, List)):
                st.write(value)
            else:
                if isinstance(value, str) and len(value) > 32:
                    st.text(value)
                else:
                    st.write(value)


def draw_llm_info(query: Lens, component: ComponentView) -> None:
    llm_details_json = component.json

    st.subheader("*LLM Details*")
    # path_str = str(query)
    # st.text(path_str[:-4])

    llm_kv = {
        k: v
        for k, v in llm_details_json.items()
        if (v is not None)
        and not is_empty(v)
        and not is_noserio(v)
        and k != CLASS_INFO
    }
    # CSS to inject contained in a string
    hide_table_row_index = """
                <style>
                thead tr th:first-child {display:none}
                tbody th {display:none}
                </style>
                """
    df = pd.DataFrame.from_dict(llm_kv, orient="index").transpose()

    # Redact any column whose name indicates it might be a secret.
    for col in df.columns:
        if should_redact_key(col):
            df[col] = REDACTED_VALUE

    # TODO: What about columns not indicating a secret but some values do
    # indicate it as per `should_redact_value` ?

    # Iterate over each column of the DataFrame
    for column in df.columns:
        path = getattr(Select.for_app(query), str(column))
        # Check if any cell in the column is a dictionary

        if any(isinstance(cell, dict) for cell in df[column]):
            # Create new columns for each key in the dictionary
            new_columns = df[column].apply(
                lambda x: pd.Series(x) if isinstance(x, dict) else pd.Series()
            )
            new_columns.columns = [
                f"{key} {render_selector_markdown(path)}"
                for key in new_columns.columns
            ]

            # Remove extra zeros after the decimal point
            new_columns = new_columns.applymap(
                lambda x: "{0:g}".format(x) if isinstance(x, float) else x
            )

            # Add the new columns to the original DataFrame
            df = pd.concat([df.drop(column, axis=1), new_columns], axis=1)

        else:
            # TODO: add selectors to the output here

            pass

    # Inject CSS with Markdown

    st.markdown(hide_table_row_index, unsafe_allow_html=True)
    st.table(df)


def draw_agent_info(query: Lens, component: ComponentView) -> None:
    # copy of draw_prompt_info
    # TODO: dedup
    prompt_details_json = jsonify(component.json, skip_specials=True)

    st.subheader("*Agent Details*")

    path = Select.for_app(query)

    prompt_types = {
        k: v
        for k, v in prompt_details_json.items()
        if (v is not None)
        and not is_empty(v)
        and not is_noserio(v)
        and k != CLASS_INFO
    }

    for key, value in prompt_types.items():
        with st.expander(
            key.capitalize()
            + " "
            + render_selector_markdown(getattr(path, key)),
            expanded=True,
        ):
            if isinstance(value, (Dict, List)):
                st.write(value)
            else:
                if isinstance(value, str) and len(value) > 32:
                    st.text(value)
                else:
                    st.write(value)


def draw_tool_info(query: Lens, component: ComponentView) -> None:
    # copy of draw_prompt_info
    # TODO: dedup
    prompt_details_json = jsonify(component.json, skip_specials=True)

    st.subheader("*Tool Details*")

    path = Select.for_app(query)

    prompt_types = {
        k: v
        for k, v in prompt_details_json.items()
        if (v is not None)
        and not is_empty(v)
        and not is_noserio(v)
        and k != CLASS_INFO
    }

    for key, value in prompt_types.items():
        with st.expander(
            key.capitalize()
            + " "
            + render_selector_markdown(getattr(path, key)),
            expanded=True,
        ):
            if isinstance(value, (Dict, List)):
                st.write(value)
            else:
                if isinstance(value, str) and len(value) > 32:
                    st.text(value)
                else:
                    st.write(value)



import asyncio
from typing import Optional

import streamlit as st
from trulens.core import TruSession
from trulens.core.schema import app as mod_app_schema
from trulens.core.schema import record as mod_record_schema
from trulens.core.utils.json import jsonify_for_ui
from trulens.core.utils.serial import JSON
from trulens.core.utils.serial import Lens
from trulens.dashboard.streamlit_utils import init_from_args
from trulens.dashboard.ux.apps import ChatRecord
from trulens.dashboard.ux.page_config import set_page_config

# https://github.com/jerryjliu/llama_index/issues/7244:
asyncio.set_event_loop(asyncio.new_event_loop())

if __name__ == "__main__":
    # If not imported, gets args from command line and creates a TruSession
    init_from_args()

session = TruSession()

set_page_config(page_title="App Runner")


def remove_selector(
    container,
    type: str,  # either "app" or "record"
    selector_idx: int,
    record_idx: int,
    selector: str,
    rec: Optional[ChatRecord] = None,
):
    """
    Remove the `selector` of type `type`. A selector should be uniquely
    addressed/keyed by `type`, `selector_idx`, and `record_idx` but don't
    presently see a reason to have duplicate selectors so indexing only by
    `type` and `selector` for now. `container` is the streamlit "empty" object
    that contains the widgets for this selector. For `record` types,
    `record_idx` is ignored as the selector is removed from all of the
    rows/records.
    """

    state = st.session_state[f"selectors_{type}"]

    if selector in state:
        state.remove(selector)
    else:
        print("no such selector")
        return

    # Get and delete all of the containers for this selector. If this is a
    # record `type`, there will be one container for each record row.
    key_norec = f"{type}_{selector_idx}"
    for container in st.session_state[f"containers_{key_norec}"]:
        del container


def update_selector(
    container,
    type: str,
    selector_idx: int,
    record_idx: int,
    selector: str,
    rec: Optional[ChatRecord] = None,
):
    """
    Update the selector keyed by `type`, `selector_idx`, `record_idx` to the new
    value retrieved from state. Only works assuming selectors are unique within
    each `type`.
    """

    state = st.session_state[f"selectors_{type}"]

    key = f"{type}_{selector_idx}_{record_idx}"

    new_val = st.session_state[f"edit_input_{key}"]

    state[state.index(selector)] = new_val


def draw_selector(
    type: str,
    selector_idx: int,
    record_idx: int,
    selector: str,
    rec: Optional[ChatRecord] = None,
):
    """
    Draws the UI elements for a selector of type `type` intended to be keyed by
    (type) and `selector_idx` and `record_idx`. The selector represents a
    Lens given as str in `selector`. Includes delete and edit widgets as
    well as the listing of the values attained by the selected path in the given
    ChatRecord `rec`.
    """

    key = f"{type}_{selector_idx}_{record_idx}"
    key_norec = f"{type}_{selector_idx}"

    container = st.empty()

    # Add the container for these elements into the state indexed by type and
    # selector so we can easily delete it later alongside its analogues in order
    # records (for "record" `type` selectors).
    if f"containers_{key_norec}" not in st.session_state:
        st.session_state[f"containers_{key_norec}"] = []
    st.session_state[f"containers_{key_norec}"].append(container)

    # Cannot stack columns too deeply:
    # c1, c2 = st.columns(2)

    # TODO: figure out how to expand/collapse these across all records at the
    # same time, this session thing does not work.
    st.session_state[f"expanded_{key_norec}"] = True

    # Put everything in this expander:
    exp = container.expander(
        label=selector, expanded=st.session_state[f"expanded_{key_norec}"]
    )

    # Edit input.
    exp.text_input(
        label="App selector",
        value=selector,
        key=f"edit_input_{key}",
        on_change=update_selector,
        kwargs=dict(
            container=container,
            type=type,
            selector_idx=selector_idx,
            record_idx=record_idx,
            selector=selector,
            rec=rec,
        ),
        label_visibility="collapsed",
    )

    # Get the relevant JSON to path into.
    obj = rec.app_json
    if type == "record":
        obj = mod_record_schema.Record.model_validate(
            rec.record_json
        ).layout_calls_as_app()

    # Try to parse the selector as a Lens.
    path = None
    try:
        path = Lens.of_string(selector)
    except Exception as e:
        exp.write(f"parse error: {e}")

    if path is not None:
        try:
            # Draw each value addressed by `path`:
            for val in path.get(obj):
                json_val = jsonify_for_ui(val)
                if isinstance(json_val, dict):
                    # Don't expand by default as there are large JSONs and
                    # streamlit scrolls to the bottom if shown.
                    exp.json(json_val, expanded=False)
                else:
                    exp.write(json_val)

        except Exception as e:
            exp.write(f"enumeration error: {e}")


def draw_rec(record_idx: int, rec: ChatRecord, skip_human: bool = False):
    """
    Draw a ChatRecord `rec`, indexed by `record_idx` into columns `cols`. If
    `skip_human` is True, skips the human part of the record (as it was already
    drawn by an earlier call).
    """

    record_json = rec.record_json

    if not skip_human:
        if rec.human is not None:
            with st.chat_message("Human", avatar="🧑‍💻"):
                st.write(rec.human)

    if rec.computer is not None:
        with st.chat_message("Computer", avatar="🤖"):
            st.write(rec.computer)

    if record_json is not None:
        # st.write(f"TODO link to {record_json['record_id']}.")

        for selector_idx, selector in enumerate(
            st.session_state.selectors_record
        ):
            draw_selector(
                type="record",
                selector_idx=selector_idx,
                record_idx=record_idx,
                selector=selector,
                rec=rec,
            )


def set_selector(type: str):
    """
    Set the selectors of type `type` in session state.
    """

    # Get value from session:
    input_key = f"set_{type}_selector_input"
    val = st.session_state[input_key]

    st.session_state[f"selectors_{type}"] = val


def add_selector(type: str):
    """
    Add a new selector of type `type`. Value is looked up from session state.
    """

    # Get value from session:
    input_key = f"add_{type}_selector_input"
    val = st.session_state[input_key]

    # The list of selectors of the appropriate type:
    state = st.session_state[f"selectors_{type}"]

    # Add selector if not already in there:
    if val not in state:
        state.append(val)

        # Clear input
        st.session_state[input_key] = ""
    else:
        print(f"{type} selector {val} already exists")


def select_app(app_json: JSON):
    """
    Select the app to start a session with by its JSON.
    """

    tru_app = mod_app_schema.AppDefinition.new_session(
        app_definition_json=app_json
    )

    st.session_state.records = [ChatRecord(app_json=app_json, app=tru_app)]

    for typ in ["app", "record"]:
        st.session_state[f"selectors_{typ}"] = []


def run_record(col):
    """
    Assuming a user provided some input in the appropriate text box, run the app
    from its final state on the new input.
    """

    # Current is last:
    last_record_index = len(st.session_state.records) - 1
    current_record = st.session_state.records[last_record_index]

    # Get the human input from state and update record.
    human_input = st.session_state["human_input"]
    current_record.human = human_input

    # Draw the ChatRecord so far, just human input.
    with col:
        draw_rec(record_idx=last_record_index, rec=current_record)

    # TODO: set some sort of progress bar or do async computation for computer
    # response.
    tru_app = current_record.app

    # Run the app and collect the record.
    with tru_app as rec:
        comp_response = tru_app.main_call(human_input)
    record: mod_record_schema.Record = rec.get()

    # Update ChatRecord.
    current_record.computer = comp_response

    # Doing this after draw_rec so that the computer output can show up before
    # we start rendering selected values.
    current_record.record_json = record.model_dump()

    # Add the next ChatRecord that contains the updated app state:
    st.session_state.records.append(
        ChatRecord(app=tru_app, app_json=tru_app.model_dump())
    )


def end_session():
    """
    Reset the state to that before a session is started.
    """

    del st.session_state["records"]


if "records" not in st.session_state:
    # This field only exists after a model is selected. Here no model was
    # selected yet. Show all loadable models for which a new session can be
    # started.

    st.title("App Runner")

    loadable_apps = mod_app_schema.AppDefinition.get_loadable_apps()

    st.divider()

    for app_json in loadable_apps:
        st.subheader(app_json["app_id"])
        st.button(
            label="New Session",
            key=f"select_app_{app_json['app_id']}",
            on_click=select_app,
            args=(app_json,),
        )

        st.divider()

    if len(loadable_apps) == 0:
        st.write(
            "No loadable apps found in database. "
            "To make an app loadable, specify a loader function via the `initial_app_loader` argument when wrapping the app. "
            "See the notebook at https://github.com/truera/trulens/blob/main/examples/experimental/streamlit_appui_example.ipynb for an example."
        )

else:
    # Otherwise a model was selected, and there should be at least one
    # ChatRecord in the state.

    first_record = st.session_state.records[0]
    app_json = first_record.app_json

    # Show the app id and some app-level or session-level controls/links.
    st.title(f"App Runner: {app_json['app_id']}")

    st.button(label="End session", on_click=end_session)

    # st.write(f"TODO: link to {app_json['app_id']} on other pages.")

    st.divider()

    left, right = st.columns([1 / 3, 2 / 3])

    with left:
        # On the left are app selectors that show the properties of the app as
        # it is at the current/final state of the session.

        st.write("#### App details")
        st.caption(
            "Details about your app. Use app selectors below to select the pieces of information to highlight."
        )

        # Create an add app selector input:
        st.write("**App selectors**")
        if len(st.session_state.selectors_app) > 0:
            st.multiselect(
                "Current app selectors",
                st.session_state.selectors_app,
                st.session_state.selectors_app,
                on_change=set_selector,
                key="set_app_selector_input",
                args=("app",),
                label_visibility="collapsed",
            )
        st.text_input(
            label="add app selector",
            key="add_app_selector_input",
            placeholder="Add an app selector (e.g. app.llm.model_name)",
            on_change=add_selector,
            args=("app",),
            label_visibility="collapsed",
        )

        # Draw existing app selectors.
        for i, selector in enumerate(st.session_state.selectors_app):
            draw_selector(
                type="app",
                selector_idx=i,
                record_idx=None,
                selector=selector,
                rec=first_record,
            )

    with right:
        # On the right 2/3 are rows, one per ChatRecord.
        st.write("#### Records")
        st.caption(
            "Your interactive chat session. Type in a message below to 'chat' with the LLM application. Record selectors can be used to surface information on a per-record level."
        )

        st.write("**Record selectors**")
        if len(st.session_state.selectors_record) > 0:
            st.multiselect(
                "Current record selectors",
                st.session_state.selectors_record,
                st.session_state.selectors_record,
                on_change=set_selector,
                key="set_record_selector_input",
                args=("record",),
                label_visibility="collapsed",
            )
        st.text_input(
            label="add record selector",
            placeholder="Add a record selector to view details about the records (e.g. cost.cost).",
            key="add_record_selector_input",
            on_change=add_selector,
            args=("record",),
            label_visibility="collapsed",
        )

        # Rows corresponding to ChatRecord:
        for i, rec in enumerate(st.session_state.records):
            draw_rec(record_idx=i, rec=rec)

        if (
            len(st.session_state.records) == 0
            or st.session_state.records[0].record_json is None
        ):
            st.caption("Begin a chat session by typing in a message below")

    # NOTE: chat input cannot be inside column.
    human_input = st.chat_input(
        on_submit=run_record,
        key="human_input",
        kwargs=dict(
            col=right  # should be the cols of the last row from the above enumeration.
        ),
    )



# Code in support of the Apps.py page.

from typing import ClassVar, Optional

import pydantic
from trulens.core import app as mod_app
from trulens.core.utils.serial import JSON


class ChatRecord(pydantic.BaseModel):
    model_config: ClassVar[dict] = dict(arbitrary_types_allowed=True)

    # Human input
    human: Optional[str] = None

    # Computer response
    computer: Optional[str] = None

    # Jsonified record. Available only after the app is run on human input and
    # produced a computer output.
    record_json: Optional[JSON] = None

    # The final app state for continuing the session.
    app: mod_app.App

    # The state of the app as was when this record was produced.
    app_json: JSON



import asyncio
import json
import math
from typing import List

from pydantic import BaseModel
import streamlit as st
from streamlit_pills import pills
from trulens.core import TruSession
from trulens.core.database.legacy.migration import MIGRATION_UNKNOWN_STR
from trulens.core.schema.feedback import FeedbackCall
from trulens.core.schema.record import Record
from trulens.core.utils.json import json_str_of_obj
from trulens.core.utils.text import format_quantity
from trulens.dashboard.components.record_viewer import record_viewer
from trulens.dashboard.display import get_feedback_result
from trulens.dashboard.display import get_icon
from trulens.dashboard.ux import styles
from trulens.dashboard.ux.components import draw_metadata_and_tags

# https://github.com/jerryjliu/llama_index/issues/7244:
asyncio.set_event_loop(asyncio.new_event_loop())


class FeedbackDisplay(BaseModel):
    score: float = 0
    calls: List[FeedbackCall]
    icon: str


def trulens_leaderboard(app_ids: List[str] = None):
    """
    Render the leaderboard page.

    Args:

        app_ids List[str]: A list of application IDs (default is None)

    Example:

        ```python
        from trulens.core import streamlit as trulens_st

        trulens_st.trulens_leaderboard()
        ```
    """
    session = TruSession()

    lms = session.connector.db
    df, feedback_col_names = lms.get_records_and_feedback(app_ids=app_ids)
    feedback_defs = lms.get_feedback_defs()
    feedback_directions = {
        (
            row.feedback_json.get("supplied_name", "")
            or row.feedback_json["implementation"]["name"]
        ): row.feedback_json.get("higher_is_better", True)
        for _, row in feedback_defs.iterrows()
    }

    if df.empty:
        st.write("No records yet...")
        return

    df.sort_values(by="app_id", inplace=True)

    if df.empty:
        st.write("No records yet...")

    def get_data():
        return lms.get_records_and_feedback([])

    def get_apps():
        return list(lms.get_apps())

    records, feedback_col_names = get_data()
    records = records.sort_values(by="app_id")

    apps = get_apps()

    for app in apps:
        app_df = records.loc[records.app_id == app]
        if app_df.empty:
            continue
        app_str = app_df["app_json"].iloc[0]
        app_json = json.loads(app_str)
        app_name = app_json["app_name"]
        app_version = app_json["app_version"]
        app_name_version = f"{app_name} - {app_version}"
        metadata = app_json.get("metadata")
        tags = app_json.get("tags")
        st.header(app_name_version, help=draw_metadata_and_tags(metadata, tags))
        app_feedback_col_names = [
            col_name
            for col_name in feedback_col_names
            if not app_df[col_name].isna().all()
        ]
        col1, col2, col3, col4, *feedback_cols = st.columns(
            5 + len(app_feedback_col_names)
        )
        latency_mean = (
            app_df["latency"]
            .apply(lambda td: td if td != MIGRATION_UNKNOWN_STR else None)
            .mean()
        )

        col1.metric("Records", len(app_df))
        col2.metric(
            "Average Latency (Seconds)",
            (
                f"{format_quantity(round(latency_mean, 5), precision=2)}"
                if not math.isnan(latency_mean)
                else "nan"
            ),
        )
        col3.metric(
            "Total Cost (USD)",
            f"${format_quantity(round(sum(cost for cost in app_df.total_cost if cost is not None), 5), precision=2)}",
        )
        col4.metric(
            "Total Tokens",
            format_quantity(
                sum(
                    tokens
                    for tokens in app_df.total_tokens
                    if tokens is not None
                ),
                precision=2,
            ),
        )

        for i, col_name in enumerate(app_feedback_col_names):
            mean = app_df[col_name].mean()

            st.write(
                styles.stmetricdelta_hidearrow,
                unsafe_allow_html=True,
            )

            higher_is_better = feedback_directions.get(col_name, True)

            if "distance" in col_name:
                feedback_cols[i].metric(
                    label=col_name,
                    value=f"{round(mean, 2)}",
                    delta_color="normal",
                )
            else:
                cat = styles.CATEGORY.of_score(
                    mean, higher_is_better=higher_is_better
                )
                feedback_cols[i].metric(
                    label=col_name,
                    value=f"{round(mean, 2)}",
                    delta=f"{cat.icon} {cat.adjective}",
                    delta_color=(
                        "normal"
                        if cat.compare(
                            mean, styles.CATEGORY.PASS[cat.direction].threshold
                        )
                        else "inverse"
                    ),
                )

        st.markdown("""---""")


@st.experimental_fragment(run_every=2)
def trulens_feedback(record: Record):
    """
    Render clickable feedback pills for a given record.

    Args:

        record (Record): A trulens record.

    Example:

        ```python
        from trulens.core import streamlit as trulens_st

        with tru_llm as recording:
            response = llm.invoke(input_text)

        record, response = recording.get()

        trulens_st.trulens_feedback(record=record)
        ```
    """
    feedback_cols = []
    feedbacks = {}
    icons = []
    for feedback, feedback_result in record.wait_for_feedback_results().items():
        call_data = {
            "feedback_definition": feedback,
            "feedback_name": feedback.name,
            "result": feedback_result.result,
        }
        feedback_cols.append(call_data["feedback_name"])
        feedbacks[call_data["feedback_name"]] = FeedbackDisplay(
            score=call_data["result"],
            calls=[],
            icon=get_icon(fdef=feedback, result=feedback_result.result),
        )
        icons.append(feedbacks[call_data["feedback_name"]].icon)

    st.write("**Feedback functions**")
    selected_feedback = pills(
        "Feedback functions",
        feedback_cols,
        index=None,
        format_func=lambda fcol: f"{fcol} {feedbacks[fcol].score:.4f}",
        label_visibility="collapsed",  # Hiding because we can't format the label here.
        icons=icons,
        key=f"{call_data['feedback_name']}_{len(feedbacks)}",  # Important! Otherwise streamlit sometimes lazily skips update even with st.fragment
    )

    if selected_feedback is not None:
        st.dataframe(
            get_feedback_result(record, feedback_name=selected_feedback),
            use_container_width=True,
            hide_index=True,
        )


def trulens_trace(record: Record):
    """
    Display the trace view for a record.

    Args:

        record (Record): A trulens record.

    Example:

        ```python
        from trulens.core import streamlit as trulens_st

        with tru_llm as recording:
            response = llm.invoke(input_text)

        record, response = recording.get()

        trulens_st.trulens_trace(record=record)
        ```
    """

    session = TruSession()
    app = session.get_app(app_id=record.app_id)
    record_viewer(record_json=json.loads(json_str_of_obj(record)), app_json=app)



import asyncio
import json
import math

import streamlit as st
from streamlit_extras.switch_page_button import switch_page
from trulens.core import TruSession
from trulens.core.database.legacy.migration import MIGRATION_UNKNOWN_STR
from trulens.core.utils.text import format_quantity
from trulens.dashboard.streamlit_utils import init_from_args
from trulens.dashboard.ux import styles
from trulens.dashboard.ux.components import draw_metadata_and_tags
from trulens.dashboard.ux.page_config import set_page_config
from trulens.dashboard.ux.styles import CATEGORY

# https://github.com/jerryjliu/llama_index/issues/7244:
asyncio.set_event_loop(asyncio.new_event_loop())

if __name__ == "__main__":
    # If not imported, gets args from command line and creates Tru singleton
    init_from_args()


def leaderboard():
    """Render the leaderboard page."""

    set_page_config(page_title="Leaderboard")

    session = TruSession()  # get singleton whether this file was imported or executed from command line.

    lms = session.connector.db

    # Set the title
    st.title("App Leaderboard")

    def get_data():
        return lms.get_records_and_feedback([])

    def get_apps():
        return list(lms.get_apps())

    # sort apps by name
    def sort_selected_apps(selected_apps, records):
        """
        Sorts the selected_apps list based on the concatenation of app_name and app_version.

        Parameters:
        selected_apps (list): List of app_ids to be sorted.
        apps (DataFrame): DataFrame containing app_id and app_json columns.

        Returns:
        list: Sorted list of selected_apps.
        """
        # Create a mapping from app_id to (app_name, app_version)
        app_info = (
            records.set_index("app_id")["app_json"]
            .apply(
                lambda x: (
                    json.loads(x).get("app_name"),
                    json.loads(x).get("app_version"),
                )
            )
            .to_dict()
        )

        # Sort selected_apps based on the concatenation of app_name and app_version
        sorted_apps = sorted(
            selected_apps,
            key=lambda app_id: f"{app_info[app_id][0]}{app_info[app_id][1]}",
        )

        return sorted_apps

    records, feedback_col_names = get_data()
    records = records.sort_values(by="app_id")

    apps = get_apps()
    app_names = sorted(list(set(app["app_name"] for app in apps)))

    selected_app_names = st.multiselect("Filter apps:", app_names, app_names)
    selected_apps = [
        app["app_id"]
        for app in apps
        if any(name in app["app_name"] for name in selected_app_names)
    ]
    st.session_state.app = selected_apps

    # Filter app versions to only include those from selected apps
    filtered_apps = [app for app in apps if app["app_id"] in selected_apps]
    app_name_versions = sorted(
        list(
            set(
                f"{app['app_name']} - {app['app_version']}"
                for app in filtered_apps
            )
        )
    )

    selected_app_name_versions = st.multiselect(
        "Filter by app version:", app_name_versions, app_name_versions
    )
    selected_apps = [
        app["app_id"]
        for app in apps
        if f"{app['app_name']} - {app['app_version']}"
        in selected_app_name_versions
    ]
    st.session_state.app = selected_apps

    with st.expander("Advanced Filters"):
        # get tag options
        tags = []
        for i in range(len(apps)):
            tags.append(apps[i]["tags"])
        unique_tags = sorted(list(set(tags)))
        # select tags
        selected_tags = st.multiselect("Filter tags:", unique_tags, unique_tags)

        # filter to apps with selected tags
        tag_selected_apps = [
            app["app_id"]
            for app in apps
            if any(tag in app["tags"] for tag in selected_tags)
        ]
        selected_apps = list(set(selected_apps) & set(tag_selected_apps))
        st.session_state.app = selected_apps

        # get metadata options
        metadata_keys_unique = set()
        for app in apps:
            metadata_keys_unique.update(app["metadata"].keys())
        metadata_keys_unique = list(metadata_keys_unique)
        # select metadata
        metadata_options = {}
        for metadata_key in metadata_keys_unique:
            unique_values = set()
            for i in range(len(apps)):
                unique_values.add(apps[i]["metadata"][metadata_key])
            metadata_options[metadata_key] = list(unique_values)

        # select metadata
        metadata_selections = metadata_options.copy()
        for metadata_key in metadata_options.keys():
            metadata_selections[metadata_key] = st.multiselect(
                "Filter " + metadata_key + ":",
                sorted(metadata_options[metadata_key]),
                sorted(metadata_options[metadata_key]),
            )

        # filter to apps with selected metadata
        metadata_selected_apps = [
            app["app_id"]
            for app in apps
            if all(
                app["metadata"][metadata_key]
                in metadata_selections[metadata_key]
                for metadata_key in metadata_selections.keys()
            )
        ]

        selected_apps = list(set(selected_apps) & set(metadata_selected_apps))
        st.session_state.app = selected_apps

    feedback_defs = lms.get_feedback_defs()
    feedback_directions = {
        (
            row.feedback_json.get("supplied_name", "")
            or row.feedback_json["implementation"]["name"]
        ): row.feedback_json.get("higher_is_better", True)
        for _, row in feedback_defs.iterrows()
    }

    if records.empty:
        st.write("No records yet...")
        return

    if records.empty:
        st.write("No records yet...")

    st.markdown("""---""")

    selected_apps = sort_selected_apps(selected_apps, records)

    for app in selected_apps:
        app_df = records.loc[records.app_id == app]
        if app_df.empty:
            continue
        app_str = app_df["app_json"].iloc[0]
        app_json = json.loads(app_str)
        app_name = app_json["app_name"]
        app_version = app_json["app_version"]
        app_name_version = f"{app_name} - {app_version}"
        metadata = app_json.get("metadata")
        tags = app_json.get("tags")
        # st.text('Metadata' + str(metadata))
        st.header(app_name_version, help=draw_metadata_and_tags(metadata, tags))
        app_feedback_col_names = [
            col_name
            for col_name in feedback_col_names
            if not app_df[col_name].isna().all()
        ]
        col1, col2, col3, col4, *feedback_cols, col99 = st.columns(
            5 + len(app_feedback_col_names)
        )
        latency_mean = (
            app_df["latency"]
            .apply(lambda td: td if td != MIGRATION_UNKNOWN_STR else None)
            .mean()
        )

        col1.metric("Records", len(app_df))
        col2.metric(
            "Average Latency (Seconds)",
            (
                f"{format_quantity(round(latency_mean, 5), precision=2)}"
                if not math.isnan(latency_mean)
                else "nan"
            ),
        )
        col3.metric(
            "Total Cost (USD)",
            f"${format_quantity(round(sum(cost for cost in app_df.total_cost if cost is not None), 5), precision=2)}",
        )
        col4.metric(
            "Total Tokens",
            format_quantity(
                sum(
                    tokens
                    for tokens in app_df.total_tokens
                    if tokens is not None
                ),
                precision=2,
            ),
        )

        for i, col_name in enumerate(app_feedback_col_names):
            mean = app_df[col_name].mean()

            st.write(
                styles.stmetricdelta_hidearrow,
                unsafe_allow_html=True,
            )

            higher_is_better = feedback_directions.get(col_name, True)

            if "distance" in col_name:
                feedback_cols[i].metric(
                    label=col_name,
                    value=f"{round(mean, 2)}",
                    delta_color="normal",
                )
            else:
                cat = CATEGORY.of_score(mean, higher_is_better=higher_is_better)
                feedback_cols[i].metric(
                    label=col_name,
                    value=f"{round(mean, 2)}",
                    delta=f"{cat.icon} {cat.adjective}",
                    delta_color=(
                        "normal"
                        if cat.compare(
                            mean, CATEGORY.PASS[cat.direction].threshold
                        )
                        else "inverse"
                    ),
                )

        with col99:
            if st.button("Select App", key=f"app-selector-{app}"):
                st.session_state.app = app
                switch_page("Evaluations")

        # with st.expander("Model metadata"):
        #    st.markdown(draw_metadata(metadata))

        st.markdown("""---""")


if __name__ == "__main__":
    leaderboard()



import time

import pandas as pd
from trulens.core.schema.feedback import FeedbackDefinition
from trulens.core.schema.record import Record
from trulens.dashboard.ux.styles import CATEGORY


def get_icon(fdef: FeedbackDefinition, result: float) -> str:
    """
    Get the icon for a given feedback definition and result.

    Args:

    fdef : FeedbackDefinition
        The feedback definition
    result : float
        The result of the feedback

    Returns:
        str: The icon for the feedback
    """
    cat = CATEGORY.of_score(
        result or 0,
        higher_is_better=fdef.higher_is_better
        if fdef.higher_is_better is not None
        else True,
    )
    return cat.icon


def get_feedback_result(
    tru_record: Record, feedback_name: str, timeout: int = 60
) -> pd.DataFrame:
    """
    Retrieve the feedback results for a given feedback name from a TruLens record.

    Args:
        tru_record: The record containing feedback and future results.

        feedback_name: The name of the feedback to retrieve results for.

    Returns:
        pd.DataFrame: A DataFrame containing the feedback results. If no feedback
                      results are found, an empty DataFrame is returned.
    """
    start_time = time.time()
    feedback_calls = None

    while time.time() - start_time < timeout:
        feedback_calls = next(
            (
                future_result.result()
                for feedback_definition, future_result in tru_record.feedback_and_future_results
                if feedback_definition.name == feedback_name
            ),
            None,
        )
        if feedback_calls is not None:
            break
        time.sleep(1)  # Wait for 1 second before checking again

    if feedback_calls is None:
        raise TimeoutError(
            f"Feedback for '{feedback_name}' not available within {timeout} seconds."
        )

    # Ensure feedback_calls is iterable
    if not hasattr(feedback_calls, "__iter__"):
        raise ValueError("feedback_calls is not iterable")

    feedback_result = [
        {**call.model_dump()["args"], "ret": call.model_dump()["ret"]}
        for call in feedback_calls.calls
    ]
    return pd.DataFrame(feedback_result)
